#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include "objz80.h"

unsigned char * workspace;
int pos = 0;

struct globalsym * symbols = NULL;
struct externsym * relocentries = NULL;
int nss = 0;
int nrle = 0;

void add_sym(unsigned char * name, unsigned short pos)
{
    symbols = (struct globalsym *) realloc(symbols, sizeof(struct globalsym)*(++nss));
    snprintf(symbols[nss-1].name, 20, name);
    symbols[nss-1].offset = pos;
}

void add_relocentry(unsigned char * name, unsigned short pos, int mode)
{
    relocentries = (struct externsym *) realloc(relocentries, sizeof(struct externsym)*(++nrle));
    relocentries[nrle-1].mode = mode;
    snprintf(relocentries[nrle-1].name, 20, name);
    relocentries[nrle-1].offset = pos;
}

int resolve(unsigned char * name)
{
    int i;
    for(i=0; i<nss; i++)	
	if(strcmp(name, symbols[i].name)==0) return(symbols[i].offset);
    return(-1);
}


void relocate(void)
{
    int q;
    int ddr, offs;
    
    printf("Relocating ...\n");
    for(q=0; q<nrle; q++)
    {
	printf("%d of %d ...\n",q,nrle);
	ddr = resolve(relocentries[q].name);
	if(ddr == -1)
	{
	    printf("*** Unresolved symbol: %s\n", relocentries[q].name);
	    exit(1);
	}
	offs = relocentries[q].offset;
/* endianness !! */
	if(relocentries[q].mode == RELOCMODE_ABSOLUTE)
	{
	    workspace[offs] = ddr & 0xff;
	    workspace[offs+1] = (ddr >> 8) & 0xff;
	}
	else
	{
            if((ddr - offs - 1) < -127 || (ddr - offs - 1) > 128)
	    {
	        fprintf(stderr,"*** Relative jump to %s too far\n", relocentries[q].name);
	        exit(0);
	    }
	    workspace[offs] = (ddr - offs - 1) & 0xff;
	}
    }
    printf("%d entries relocated succefully\n",nrle);
}

void help(void)
{
    printf("mteg's z80 linker for mteg's z80 assembler\n");
    printf("written 2001, licensed under GNU General Public License\n");
    printf("\nusage:\n  linkz80 [-r outfile] [-m memory start] infile1 infile2 infile3 ...\n\n");
}

int main(int argc, char ** argv)
{
    int opt;
    char * romimg = "z80.rom";
    char * infile;
    unsigned char * inmem;
    int fd, flen, i;
    long int base = 0;
    
    for(;;)
    {
	opt=getopt(argc,argv,"hr:m:");
        if(opt==EOF) break;
	switch(opt)
	{
	    case 'h':
		help();
		exit(0);
		break;
	    case 'r': if(optarg) romimg=strdup(optarg); break;
	    case 'm': base = strtol(optarg, NULL, 0); break;
	}
    }
    
    if(!(infile = argv[optind]))
    {
	fprintf(stderr,"No input files specified ! \n");
	exit(1);
    }

    if((fd = open(romimg, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IROTH|S_IRGRP)) == -1)
    {
	fprintf(stderr,"Unable to open the output file!\n");
	exit(1);
    }
    
    workspace = (unsigned char*) malloc(65536);
    
    printf("Loading objects into memory ...\n");
    for(i = optind; argv[i]; i++)
    {
	struct z80fileinfo * fi;
	unsigned char * obj;
	int n,j;
	unsigned char * orig;
	
	/* load file */
	printf("%s: Load\n",argv[i]);
        if(!(orig = inmem = load(argv[i],&flen)))
	{
	    fprintf(stderr,"Unable to load '%s'!\n",argv[i]);
	    exit(1);
	}
	fi = (struct z80fileinfo*) inmem;
	/* scan header */
	printf("%s: Identify\n",argv[i]);
	if(fi->header != Z80_MAGIC)
	{
	    fprintf(stderr,"File '%s' not in z80o format!\n",argv[i]);
	    exit(1);
	}
	obj = inmem + sizeof(struct z80fileinfo) + (fi->relocentries * sizeof(struct relocentry)) + (fi->globals * sizeof(struct globalsym)) + (fi->externs * sizeof(struct externsym));
	
	inmem += sizeof(struct z80fileinfo);
	/* relocate code */
	printf("%s: Move orgin\n",argv[i]);
	for(j=0; j<fi->relocentries; j++)
	{
	    struct relocentry * re = (struct relocentry*) inmem;
	    int ptr, addr;
	    ptr = re->offset;
	    addr = (obj[ptr+1] << 8) + obj[ptr];
	    addr += pos + base;
	    obj[ptr] = addr & 0xff;
	    obj[ptr+1] = (addr >> 8) & 0xff;
	    
	    inmem += sizeof(struct relocentry);
	}
	printf("%s: Scan references\n",argv[i]);
	for(j=0; j<fi->externs; j++)
	{
	    struct externsym * ex = (struct externsym*) inmem;
	    add_relocentry(ex->name, ex->offset + pos, ex->mode);
	    inmem += sizeof(struct externsym);
	}

	printf("%s: Scan globals\n",argv[i]);
	for(j=0; j<fi->globals; j++)
	{
	    struct globalsym * gl = (struct globalsym*) inmem;
	    add_sym(gl->name, gl->offset + pos + base);
	    inmem += sizeof(struct globalsym);
	}
	
	printf("%s: Link\n",argv[i]);
	memcpy(workspace+pos, obj, fi->code_length);
	pos+=fi->code_length;
	
	free(orig);
	
    }
    relocate();
    write(fd, workspace, pos);
    close(fd);
    return(0);
}
